@@ -1,13 +1,12 @@
 /* See LICENSE file for copyright and license details. */
-#include <stdbool.h>
 #include <ctype.h>
 #include <errno.h>
 #include <inttypes.h>
 #include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
-#include <stdlib.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
 #include "util.h"
@@ -21,12 +20,13 @@ struct properties_payload {
 	struct properties *prop;
 	const struct property_spec *spec;
 	uint_least8_t speclen;
-	int (*set_value)(struct properties_payload *, uint_least32_t, int_least64_t);
-	uint_least8_t (*handle_conflict)(uint_least32_t, uint_least8_t, uint_least8_t);
+	int (*set_value)(struct properties_payload *, uint_least32_t,
+	                 int_least64_t);
+	uint_least8_t (*handle_conflict)(uint_least32_t, uint_least8_t,
+	                                 uint_least8_t);
 };
 
-struct break_test_payload
-{
+struct break_test_payload {
 	struct break_test **test;
 	size_t *testlen;
 };
@@ -51,8 +51,8 @@ hextocp(const char *str, size_t len, uint_least32_t *cp)
 
 	/* the maximum valid codepoint is 0x10FFFF */
 	if (len > 6) {
-		fprintf(stderr, "hextocp: '%.*s' is too long.\n",
-		        (int)len, str);
+		fprintf(stderr, "hextocp: '%.*s' is too long.\n", (int)len,
+		        str);
 		return 1;
 	}
 
@@ -77,14 +77,53 @@ hextocp(const char *str, size_t len, uint_least32_t *cp)
 	}
 
 	if (*cp > UINT32_C(0x10FFFF)) {
-		fprintf(stderr, "hextocp: '%.*s' is too large.\n",
-		        (int)len, str);
+		fprintf(stderr, "hextocp: '%.*s' is too large.\n", (int)len,
+		        str);
 		return 1;
 	}
 
 	return 0;
 }
 
+int
+parse_cp_list(const char *str, uint_least32_t **cp, size_t *cplen)
+{
+	size_t count, i;
+	const char *tmp1 = NULL, *tmp2 = NULL;
+
+	if (strlen(str) == 0) {
+		*cp = NULL;
+		*cplen = 0;
+		return 0;
+	}
+
+	/* count the number of spaces in the string and infer list length */
+	for (count = 1, tmp1 = str; (tmp2 = strchr(tmp1, ' ')) != NULL;
+	     count++, tmp1 = tmp2 + 1) {
+		;
+	}
+
+	/* allocate resources */
+	if (!(*cp = calloc((*cplen = count), sizeof(**cp)))) {
+		fprintf(stderr, "calloc: %s\n", strerror(errno));
+		exit(1);
+	}
+
+	/* go through the string again, parsing the numbers */
+	for (i = 0, tmp1 = tmp2 = str; tmp2 != NULL; i++) {
+		tmp2 = strchr(tmp1, ' ');
+		if (hextocp(tmp1, tmp2 ? (size_t)(tmp2 - tmp1) : strlen(tmp1),
+		            &((*cp)[i]))) {
+			return 1;
+		}
+		if (tmp2 != NULL) {
+			tmp1 = tmp2 + 1;
+		}
+	}
+
+	return 0;
+}
+
 static int
 range_parse(const char *str, struct range *range)
 {
@@ -107,14 +146,52 @@ range_parse(const char *str, struct range *range)
 	return 0;
 }
 
+static bool
+get_line(char **buf, size_t *bufsize, FILE *fp, size_t *len)
+{
+	int ret = EOF;
+
+	for (*len = 0;; (*len)++) {
+		if (*len > 0 && *buf != NULL && (*buf)[*len - 1] == '\n') {
+			/*
+			 * if the previously read character was a newline,
+			 * we fake an end-of-file so we NUL-terminate and
+			 * are done.
+			 */
+			ret = EOF;
+		} else {
+			ret = fgetc(fp);
+		}
+
+		if (*len >= *bufsize) {
+			/* the buffer needs to be expanded */
+			*bufsize += 512;
+			if ((*buf = realloc(*buf, *bufsize)) == NULL) {
+				fprintf(stderr, "get_line: Out of memory.\n");
+				exit(1);
+			}
+		}
+
+		if (ret != EOF) {
+			(*buf)[*len] = (char)ret;
+		} else {
+			(*buf)[*len] = '\0';
+			break;
+		}
+	}
+
+	return *len == 0 && (feof(fp) || ferror(fp));
+}
+
 void
-parse_file_with_callback(const char *fname, int (*callback)(const char *,
-                         char **, size_t, char *, void *), void *payload)
+parse_file_with_callback(const char *fname,
+                         int (*callback)(const char *, char **, size_t, char *,
+                                         void *),
+                         void *payload)
 {
 	FILE *fp;
 	char *line = NULL, **field = NULL, *comment;
-	size_t linebufsize = 0, i, fieldbufsize = 0, j, nfields;
-	ssize_t len;
+	size_t linebufsize = 0, i, fieldbufsize = 0, j, nfields, len;
 
 	/* open file */
 	if (!(fp = fopen(fname, "r"))) {
@@ -123,7 +200,7 @@ parse_file_with_callback(const char *fname, int (*callback)(const char *,
 		exit(1);
 	}
 
-	while ((len = getline(&line, &linebufsize, fp)) >= 0) {
+	while (!get_line(&line, &linebufsize, fp, &len)) {
 		/* remove trailing newline */
 		if (len > 0 && line[len - 1] == '\n') {
 			line[len - 1] = '\0';
@@ -146,10 +223,15 @@ parse_file_with_callback(const char *fname, int (*callback)(const char *,
 			if (line[i] != '#') {
 				/* extend field buffer, if necessary */
 				if (++nfields > fieldbufsize) {
-					if ((field = realloc(field, nfields *
-					      sizeof(*field))) == NULL) {
-						fprintf(stderr, "parse_file_with_"
-						        "callback: realloc: %s.\n",
+					if ((field = realloc(
+						     field,
+						     nfields *
+							     sizeof(*field))) ==
+					    NULL) {
+						fprintf(stderr,
+						        "parse_file_with_"
+						        "callback: realloc: "
+						        "%s.\n",
 						        strerror(errno));
 						exit(1);
 					}
@@ -173,8 +255,9 @@ parse_file_with_callback(const char *fname, int (*callback)(const char *,
 
 			/* go back whitespace and terminate field there */
 			if (i > 0) {
-				for (j = i - 1; line[j] == ' '; j--)
+				for (j = i - 1; line[j] == ' '; j--) {
 					;
+				}
 				line[j + 1] = '\0';
 			} else {
 				line[i] = '\0';
@@ -194,11 +277,19 @@ parse_file_with_callback(const char *fname, int (*callback)(const char *,
 		/* call callback function */
 		if (callback(fname, field, nfields, comment, payload)) {
 			fprintf(stderr, "parse_file_with_callback: "
-			        "Malformed input.\n");
+			                "Malformed input.\n");
 			exit(1);
 		}
 	}
 
+	/* close file */
+	if (fclose(fp)) {
+		fprintf(stderr, "parse_file_with_callback: fclose '%s': %s.\n",
+		        fname, strerror(errno));
+		exit(1);
+	}
+
+	/* cleanup */
 	free(line);
 	free(field);
 }
@@ -221,11 +312,15 @@ properties_callback(const char *file, char **field, size_t nfields,
 
 	for (i = 0; i < p->speclen; i++) {
 		/* identify fitting file and identifier */
-		if (p->spec[i].file &&
-		    !strcmp(p->spec[i].file, file) &&
+		if (p->spec[i].file && !strcmp(p->spec[i].file, file) &&
 		    (!strcmp(p->spec[i].ucdname, field[1]) ||
-		     (comment != NULL && !strncmp(p->spec[i].ucdname, comment, strlen(p->spec[i].ucdname)) &&
-		      comment[strlen(p->spec[i].ucdname)] == ' '))) {
+		     (comment != NULL &&
+		      !strncmp(p->spec[i].ucdname, comment,
+		               strlen(p->spec[i].ucdname)) &&
+		      comment[strlen(p->spec[i].ucdname)] == ' ')) &&
+		    (p->spec[i].ucdsubname == NULL ||
+		     (nfields >= 3 &&
+		      !strcmp(p->spec[i].ucdsubname, field[2])))) {
 			/* parse range in first field */
 			if (range_parse(field[0], &r)) {
 				return 1;
@@ -251,7 +346,8 @@ properties_compress(const struct properties *prop,
 	uint_least32_t cp, i;
 
 	/* initialization */
-	if (!(comp->offset = malloc((size_t)UINT32_C(0x110000) * sizeof(*(comp->offset))))) {
+	if (!(comp->offset = malloc((size_t)UINT32_C(0x110000) *
+	                            sizeof(*(comp->offset))))) {
 		fprintf(stderr, "malloc: %s\n", strerror(errno));
 		exit(1);
 	}
@@ -260,7 +356,8 @@ properties_compress(const struct properties *prop,
 
 	for (cp = 0; cp < UINT32_C(0x110000); cp++) {
 		for (i = 0; i < comp->datalen; i++) {
-			if (!memcmp(&(prop[cp]), &(comp->data[i]), sizeof(*prop))) {
+			if (!memcmp(&(prop[cp]), &(comp->data[i]),
+			            sizeof(*prop))) {
 				/* found a match! */
 				comp->offset[cp] = i;
 				break;
@@ -272,9 +369,9 @@ properties_compress(const struct properties *prop,
 			 * add current properties to data and add the
 			 * offset in the offset-table
 			 */
-			if (!(comp->data = reallocate_array(comp->data,
-			                                    ++(comp->datalen),
-			                                    sizeof(*(comp->data))))) {
+			if (!(comp->data = reallocate_array(
+				      comp->data, ++(comp->datalen),
+				      sizeof(*(comp->data))))) {
 				fprintf(stderr, "reallocate_array: %s\n",
 				        strerror(errno));
 				exit(1);
@@ -321,8 +418,7 @@ properties_get_major_minor(const struct properties_compressed *comp,
 		 * and need less storage)
 		 */
 		for (j = 0; j + 0xFF < mm->minorlen; j++) {
-			if (!memcmp(&(comp->offset[i << 8]),
-			            &(mm->minor[j]),
+			if (!memcmp(&(comp->offset[i << 8]), &(mm->minor[j]),
 			            sizeof(*(comp->offset)) * 0x100)) {
 				break;
 			}
@@ -337,9 +433,9 @@ properties_get_major_minor(const struct properties_compressed *comp,
 			 * in major
 			 */
 			mm->minorlen += 0x100;
-			if (!(mm->minor = reallocate_array(mm->minor,
-			                                   mm->minorlen,
-			                                   sizeof(*(mm->minor))))) {
+			if (!(mm->minor =
+			              reallocate_array(mm->minor, mm->minorlen,
+			                               sizeof(*(mm->minor))))) {
 				fprintf(stderr, "reallocate_array: %s\n",
 				        strerror(errno));
 				exit(1);
@@ -356,9 +452,10 @@ properties_get_major_minor(const struct properties_compressed *comp,
 }
 
 void
-properties_print_lookup_table(char *name, size_t *data, size_t datalen)
+properties_print_lookup_table(const char *name, const size_t *data,
+                              size_t datalen)
 {
-	char *type;
+	const char *type;
 	size_t i, maxval;
 
 	for (i = 0, maxval = 0; i < datalen; i++) {
@@ -367,7 +464,7 @@ properties_print_lookup_table(char *name, size_t *data, size_t datalen)
 		}
 	}
 
-	type = (maxval <= UINT_LEAST8_MAX)  ? "uint_least8_t"  :
+	type = (maxval <= UINT_LEAST8_MAX)  ? "uint_least8_t" :
 	       (maxval <= UINT_LEAST16_MAX) ? "uint_least16_t" :
 	       (maxval <= UINT_LEAST32_MAX) ? "uint_least32_t" :
 	                                      "uint_least64_t";
@@ -382,21 +479,51 @@ properties_print_lookup_table(char *name, size_t *data, size_t datalen)
 		} else {
 			printf(",\n\t");
 		}
-
 	}
 	printf("};\n");
 }
 
 void
-properties_print_derived_lookup_table(char *name, char *type, size_t *offset, size_t offsetlen,
-                                      int_least64_t (*get_value)(const struct properties *,
-                                      size_t), const void *payload)
+properties_print_derived_lookup_table(
+	char *name, size_t *offset, size_t offsetlen,
+	int_least64_t (*get_value)(const struct properties *, size_t),
+	const void *payload)
 {
+	const char *type;
 	size_t i;
+	int_least64_t minval, maxval;
+
+	for (i = 0, minval = INT_LEAST64_MAX, maxval = INT_LEAST64_MIN;
+	     i < offsetlen; i++) {
+		if (get_value(payload, offset[i]) > maxval) {
+			maxval = get_value(payload, offset[i]);
+		} else if (get_value(payload, offset[i]) < minval) {
+			minval = get_value(payload, offset[i]);
+		}
+	}
+
+	if (minval < 0) {
+		/* we need a signed type */
+		type = (minval >= INT_LEAST8_MIN && maxval <= INT_LEAST8_MAX) ?
+		               "int_least8_t" :
+		       (minval >= INT_LEAST16_MIN &&
+		        maxval <= INT_LEAST16_MAX) ?
+		               "int_least16_t" :
+		       (minval >= INT_LEAST32_MIN &&
+		        maxval <= INT_LEAST32_MAX) ?
+		               "int_least32_t" :
+		               "int_least64_t";
+	} else {
+		/* we are fine with an unsigned type */
+		type = (maxval <= UINT_LEAST8_MAX)  ? "uint_least8_t" :
+		       (maxval <= UINT_LEAST16_MAX) ? "uint_least16_t" :
+		       (maxval <= UINT_LEAST32_MAX) ? "uint_least32_t" :
+		                                      "uint_least64_t";
+	}
 
 	printf("static const %s %s[] = {\n\t", type, name);
 	for (i = 0; i < offsetlen; i++) {
-		printf("%"PRIiLEAST64, get_value(payload, offset[i]));
+		printf("%" PRIiLEAST64, get_value(payload, offset[i]));
 		if (i + 1 == offsetlen) {
 			printf("\n");
 		} else if ((i + 1) % 8 != 0) {
@@ -404,7 +531,6 @@ properties_print_derived_lookup_table(char *name, char *type, size_t *offset, si
 		} else {
 			printf(",\n\t");
 		}
-
 	}
 	printf("};\n");
 }
@@ -426,19 +552,21 @@ static int
 set_value_bp(struct properties_payload *payload, uint_least32_t cp,
              int_least64_t value)
 {
-	if (payload->prop[cp].property != 0) {
+	if (payload->prop[cp].property != payload->speclen) {
 		if (payload->handle_conflict == NULL) {
-			fprintf(stderr, "set_value_bp: "
-	                        "Unhandled character break property "
+			fprintf(stderr,
+			        "set_value_bp: "
+			        "Unhandled character break property "
 			        "overwrite for 0x%06X (%s <- %s).\n",
-		                cp, payload->spec[payload->prop[cp].
-			        property].enumname,
+			        cp,
+			        payload->spec[payload->prop[cp].property]
+			                .enumname,
 			        payload->spec[value].enumname);
 			return 1;
 		} else {
-			value = payload->handle_conflict(cp,
-			        (uint_least8_t)payload->prop[cp].property,
-			        (uint_least8_t)value);
+			value = payload->handle_conflict(
+				cp, (uint_least8_t)payload->prop[cp].property,
+				(uint_least8_t)value);
 		}
 	}
 	payload->prop[cp].property = value;
@@ -449,17 +577,17 @@ set_value_bp(struct properties_payload *payload, uint_least32_t cp,
 static int_least64_t
 get_value_bp(const struct properties *prop, size_t offset)
 {
-	return (uint_least8_t)prop[offset].property;
+	return prop[offset].property;
 }
 
 void
-properties_generate_break_property(const struct property_spec *spec,
-                                   uint_least8_t speclen,
-                                   uint_least8_t (*handle_conflict)(
-                                   uint_least32_t, uint_least8_t,
-                                   uint_least8_t), uint_least8_t
-                                   (*post_process)(uint_least8_t),
-                                   const char *prefix, const char *argv0)
+properties_generate_break_property(
+	const struct property_spec *spec, uint_least8_t speclen,
+	uint_least8_t (*fill_missing)(uint_least32_t),
+	uint_least8_t (*handle_conflict)(uint_least32_t, uint_least8_t,
+                                         uint_least8_t),
+	void (*post_process)(struct properties *), const char *prefix,
+	const char *argv0)
 {
 	struct properties_compressed comp;
 	struct properties_major_minor mm;
@@ -468,11 +596,17 @@ properties_generate_break_property(const struct property_spec *spec,
 	size_t i, j, prefixlen = strlen(prefix);
 	char buf1[64], prefix_uc[64], buf2[64], buf3[64], buf4[64];
 
-	/* allocate property buffer for all 0x110000 codepoints */
+	/*
+	 * allocate property buffer for all 0x110000 codepoints and
+	 * initialize its entries to the known invalid value "speclen"
+	 */
 	if (!(prop = calloc(UINT32_C(0x110000), sizeof(*prop)))) {
 		fprintf(stderr, "calloc: %s\n", strerror(errno));
 		exit(1);
 	}
+	for (i = 0; i < UINT32_C(0x110000); i++) {
+		prop[i].property = speclen;
+	}
 
 	/* generate data */
 	payload.prop = prop;
@@ -493,28 +627,38 @@ properties_generate_break_property(const struct property_spec *spec,
 		if (i == j && spec[i].file) {
 			/* file has not been processed yet */
 			parse_file_with_callback(spec[i].file,
-			                         properties_callback,
-			                         &payload);
+			                         properties_callback, &payload);
+		}
+	}
+
+	/* fill in the missing properties that weren't explicitly given */
+	for (i = 0; i < UINT32_C(0x110000); i++) {
+		if (payload.prop[i].property == speclen) {
+			if (fill_missing != NULL) {
+				payload.prop[i].property =
+					fill_missing((uint_least32_t)i);
+			} else {
+				payload.prop[i].property = 0;
+			}
 		}
 	}
 
 	/* post-processing */
 	if (post_process != NULL) {
-		for (i = 0; i < UINT32_C(0x110000); i++) {
-			payload.prop[i].property =
-				post_process((uint_least8_t)payload.prop[i].property);
-		}
+		post_process(payload.prop);
 	}
 
 	/* compress data */
-	printf("/* Automatically generated by %s */\n#include <stdint.h>\n\n", argv0);
+	printf("/* Automatically generated by %s */\n#include <stdint.h>\n\n",
+	       argv0);
 	properties_compress(prop, &comp);
 
-	fprintf(stderr, "%s: %s-LUT compression-ratio: %.2f%%\n", argv0,
-	        prefix, properties_get_major_minor(&comp, &mm));
+	fprintf(stderr, "%s: %s-LUT compression-ratio: %.2f%%\n", argv0, prefix,
+	        properties_get_major_minor(&comp, &mm));
 
 	/* prepare names */
-	if ((size_t)snprintf(buf1, LEN(buf1), "%s_property", prefix) >= LEN(buf1)) {
+	if ((size_t)snprintf(buf1, LEN(buf1), "%s_property", prefix) >=
+	    LEN(buf1)) {
 		fprintf(stderr, "snprintf: String truncated.\n");
 		exit(1);
 	}
@@ -526,9 +670,12 @@ properties_generate_break_property(const struct property_spec *spec,
 		prefix_uc[i] = (char)toupper(prefix[i]);
 	}
 	prefix_uc[prefixlen] = '\0';
-	if ((size_t)snprintf(buf2, LEN(buf2), "%s_PROP", prefix_uc) >= LEN(buf2) ||
-	    (size_t)snprintf(buf3, LEN(buf3), "%s_major", prefix) >= LEN(buf3)   ||
-	    (size_t)snprintf(buf4, LEN(buf4), "%s_minor", prefix) >= LEN(buf4)) {
+	if ((size_t)snprintf(buf2, LEN(buf2), "%s_PROP", prefix_uc) >=
+	            LEN(buf2) ||
+	    (size_t)snprintf(buf3, LEN(buf3), "%s_major", prefix) >=
+	            LEN(buf3) ||
+	    (size_t)snprintf(buf4, LEN(buf4), "%s_minor", prefix) >=
+	            LEN(buf4)) {
 		fprintf(stderr, "snprintf: String truncated.\n");
 		exit(1);
 	}
@@ -537,7 +684,7 @@ properties_generate_break_property(const struct property_spec *spec,
 	properties_print_enum(spec, speclen, buf1, buf2);
 	properties_print_lookup_table(buf3, mm.major, 0x1100);
 	printf("\n");
-	properties_print_derived_lookup_table(buf4, "uint_least8_t", mm.minor, mm.minorlen,
+	properties_print_derived_lookup_table(buf4, mm.minor, mm.minorlen,
 	                                      get_value_bp, comp.data);
 
 	/* free data */
@@ -550,11 +697,12 @@ properties_generate_break_property(const struct property_spec *spec,
 
 static int
 break_test_callback(const char *fname, char **field, size_t nfields,
-                      char *comment, void *payload)
+                    char *comment, void *payload)
 {
 	struct break_test *t,
 		**test = ((struct break_test_payload *)payload)->test;
-	size_t i, *testlen = ((struct break_test_payload *)payload)->testlen;
+	size_t i, *testlen = ((struct break_test_payload *)payload)->testlen,
+		  commentlen;
 	char *token;
 
 	(void)fname;
@@ -573,42 +721,50 @@ break_test_callback(const char *fname, char **field, size_t nfields,
 	memset(t, 0, sizeof(*t));
 
 	/* parse testcase "<÷|×> <cp> <÷|×> ... <cp> <÷|×>" */
-	for (token = strtok(field[0], " "), i = 0; token != NULL; i++,
-	     token = strtok(NULL, " ")) {
+	for (token = strtok(field[0], " "), i = 0; token != NULL;
+	     i++, token = strtok(NULL, " ")) {
 		if (i % 2 == 0) {
 			/* delimiter or start of sequence */
-			if (i == 0 || !strncmp(token, "\xC3\xB7", 2)) { /* UTF-8 */
+			if (i == 0 ||
+			    !strncmp(token, "\xC3\xB7", 2)) { /* UTF-8 */
 				/*
 				 * '÷' indicates a breakpoint,
 				 * the current length is done; allocate
 				 * a new length field and set it to 0
 				 */
-				if ((t->len = realloc(t->len,
-				     ++t->lenlen * sizeof(*t->len))) == NULL) {
-					fprintf(stderr, "break_test_"
+				if ((t->len = realloc(
+					     t->len,
+					     ++t->lenlen * sizeof(*t->len))) ==
+				    NULL) {
+					fprintf(stderr,
+					        "break_test_"
 					        "callback: realloc: %s.\n",
 					        strerror(errno));
 					return 1;
 				}
 				t->len[t->lenlen - 1] = 0;
 			} else if (!strncmp(token, "\xC3\x97", 2)) { /* UTF-8 */
-				/*
-				 * '×' indicates a non-breakpoint, do nothing
-				 */
+				/* '×' indicates a non-breakpoint, do nothing */
 			} else {
-				fprintf(stderr, "break_test_callback: "
-				        "Malformed delimiter '%s'.\n", token);
+				fprintf(stderr,
+				        "break_test_callback: "
+				        "Malformed delimiter '%s'.\n",
+				        token);
 				return 1;
 			}
 		} else {
 			/* add codepoint to cp-array */
-			if ((t->cp = realloc(t->cp, ++t->cplen *
-			                     sizeof(*t->cp))) == NULL) {
-				fprintf(stderr, "break_test_callback: "
-				        "realloc: %s.\n", strerror(errno));
+			if ((t->cp = realloc(t->cp,
+			                     ++t->cplen * sizeof(*t->cp))) ==
+			    NULL) {
+				fprintf(stderr,
+				        "break_test_callback: "
+				        "realloc: %s.\n",
+				        strerror(errno));
 				return 1;
 			}
-			if (hextocp(token, strlen(token), &t->cp[t->cplen - 1])) {
+			if (hextocp(token, strlen(token),
+			            &t->cp[t->cplen - 1])) {
 				return 1;
 			}
 			if (t->lenlen > 0) {
@@ -616,7 +772,7 @@ break_test_callback(const char *fname, char **field, size_t nfields,
 			}
 		}
 	}
-	if (t->len[t->lenlen - 1] == 0) {
+	if (t->lenlen > 0 && t->len[t->lenlen - 1] == 0) {
 		/*
 		 * we allocated one more length than we needed because
 		 * the breakpoint was at the end
@@ -625,18 +781,22 @@ break_test_callback(const char *fname, char **field, size_t nfields,
 	}
 
 	/* store comment */
-	if (((*test)[*testlen - 1].descr = strdup(comment)) == NULL) {
-		fprintf(stderr, "break_test_callback: strdup: %s.\n",
-		        strerror(errno));
-		return 1;
+	if (comment != NULL) {
+		commentlen = strlen(comment) + 1;
+		if (((*test)[*testlen - 1].descr = malloc(commentlen)) ==
+		    NULL) {
+			fprintf(stderr, "break_test_callback: malloc: %s.\n",
+			        strerror(errno));
+			return 1;
+		}
+		memcpy((*test)[*testlen - 1].descr, comment, commentlen);
 	}
 
 	return 0;
 }
 
 void
-break_test_list_parse(char *fname, struct break_test **test,
-                        size_t *testlen)
+break_test_list_parse(char *fname, struct break_test **test, size_t *testlen)
 {
 	struct break_test_payload pl = {
 		.test = test,
@@ -650,13 +810,14 @@ break_test_list_parse(char *fname, struct break_test **test,
 
 void
 break_test_list_print(const struct break_test *test, size_t testlen,
-                        const char *identifier, const char *progname)
+                      const char *identifier, const char *progname)
 {
 	size_t i, j;
 
 	printf("/* Automatically generated by %s */\n"
 	       "#include <stdint.h>\n#include <stddef.h>\n\n"
-	       "#include \"../gen/types.h\"\n\n", progname);
+	       "#include \"../gen/types.h\"\n\n",
+	       progname);
 
 	printf("static const struct break_test %s[] = {\n", identifier);
 	for (i = 0; i < testlen; i++) {
